package R2D2Bot; # $Id: R2D2Bot.pm 248 2021-11-16 17:26:22Z minus $
use strict;
use utf8;
use feature qw/ say /;

=encoding utf-8

=head1 NAME

R2D2Bot - Telegram Bot for system notifications

=head1 VERSION

Version 1.01

=head1 SYNOPSIS

    use R2D2Bot;

=head1 DESCRIPTION

Telegram Bot for system notifications

=head2 again

Internal use only

=head2 client

    my $client = $app->client;

Returns the Mojo client (user agent) instance

=head2 execmd

    my %exest = $app->execmd("command", "arg1", "arg2", "argn");

Performs execute system commands and returns hash:

=over 8

=item command

The command line

=item status

The status of operation. 1 - no errors; 0 - error

=item exitval

The exitval value

=item error

The error message

=item output

The data from program

=back

=head2 getcam

    my $cmd_arr = $self->getcam("foo");

Get cam command by name

=head2 raise

    $app->raise("Error string");

Prints error string to STDERR pipe

=head2 FUNCTIONS

=over

=item B<explain>

    print explain( $object );

Returns Data::Dumper dump

=item B<blue>, B<cyan>, B<green>, B<red>, B<yellow>, B<magenta>

    print cyan("Format %s", "text");

Returns colored string

=back

=head1 HISTORY

See C<Changes> file

=head1 TO DO

See C<TODO> file

=head1 SEE ALSO

L<Mojolicious>, L<CTK>

=head1 AUTHOR

Serż Minus (Sergey Lepenkov) L<http://www.serzik.com> E<lt>minus@serzik.comE<gt>

=head1 COPYRIGHT

Copyright (C) 1998-2021 D&D Corporation. All Rights Reserved

=head1 LICENSE

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

See C<LICENSE> file

=cut

use vars qw/ $VERSION /;
$VERSION = "1.01";

use Carp;
use Encode; # Encode::_utf8_on();
use Encode::Locale;

use Cwd qw/getcwd/;
use Data::Dumper; #$Data::Dumper::Deparse = 1;
use File::stat qw//;
use POSIX qw//;
use Term::ANSIColor qw/colored/;
use Text::ParseWords qw/ shellwords /;

use Mojo::File qw/path/;
use Mojo::Util qw/tablify steady_time/;
use Mojo::Server::Prefork;

use CTK::Util qw/preparedir dtf tz_diff isTrueFlag rundir sharedir sharedstatedir trim/;
use CTK::TFVals qw/ :ALL /;
use CTK::ConfGenUtil;

use R2D2Bot::Server;
use R2D2Bot::Client;

use base qw/ CTK::App /;

use constant {
        PROJECTNAME         => "R2D2Bot",
        PROJECTNAMEL        => "r2d2bot",
        DATADIR             => "r2d2bot",
        IS_TTY              => (-t STDOUT) ? 1 : 0,
        IS_ROOT             => ($> == 0) ? 1 : 0,

        # Server
        SERVER_URL          => 'https://localhost:7232/bot',
        SERVER_LISTEN_PORT  => 7232, # r2d2
        SERVER_LISTEN_ADDR  => "*",

        # UID/GID for daemon
        SERVER_USERNAME     => "artoo-detoo",
        SERVER_GROUPNAME    => "artoo-detoo",

        UPGRADE_TIMEOUT     => 30,
    };

__PACKAGE__->register_handler(
    handler     => "version",
    description => sprintf("%s Version", PROJECTNAME),
    code => sub {
### CODE:
    my ($self, $meta, @arguments) = @_;
    printf("%s/%s\n", PROJECTNAME, $self->VERSION);
    return 1;
});

__PACKAGE__->register_handler(
    handler     => "status",
    description => "Get status information",
    code => sub {
### CODE:
    my ($self, $meta, @arguments) = @_;

    printf("Version         : %s\n", $self->VERSION);
    printf("Data dir        : %s\n", $self->datadir);
    printf("Temp dir        : %s\n", $self->tempdir);
    printf("Config file     : %s\n", $self->conf("loadstatus") ? $self->configfile : yellow("not loaded"));
        $self->debug(explain($self->config)) if $self->conf("loadstatus") && $self->verbosemode;

    # API Token
    my $token = value($self->conf("apitoken"));
    printf("API Token       : %s\n", $token ? $token : red("no token"));
    #$self->raise("The fingerprint is incorrect");

    # Proxy
    my $proxy = value($self->conf("proxy"));
    printf("Proxy server    : %s\n", $proxy) if $proxy;

    # WebHook
    my $webhook = value($self->conf("webhook"));
    printf("WebHook URL     : %s\n", $webhook ? $webhook : yellow("no webhook"));

    # Server
    my $client = $self->client;
    $client->check(); # Check
    printf("Server URL      : %s\n", $client->url ? $client->url->to_string : yellow("unknown"));
    printf("Server status   : %s\n", $client->status ? green("ok") : red($client->error));
    $self->debug($client->trace);
    $self->debug($client->status ? explain($client->res->json) : $client->res->body) if $self->verbosemode;

    return 1;
});

__PACKAGE__->register_handler(
    handler     => "webhook",
    description => "Set webhook URL for bot",
    code => sub {
### CODE:
    my ($self, $meta, @arguments) = @_;
    my $webhook = shift(@arguments) || value($self->conf("webhook"));

    # Check webhook
    return $self->raise("No webhook specified. Check configuration file") unless $webhook;

    # Check server
    my $client = $self->client;
    unless ($client->check()) {
        $self->debug($client->trace);
        $self->debug($client->res->body) if $self->verbosemode;
        return $self->raise($client->error);
    }

    # Get webhook info
    $client->request(GET => "getWebhookInfo") or do {
        $self->debug($client->trace);
        $self->debug($client->res->body) if $self->verbosemode;
        return $self->raise($client->error);
    };
    my $url = $client->res->json("/result/url");
    if ($url && $url eq $webhook) { # Skip
        say yellow("Skipped! No change required. Webhook URL: %s", $url);
        return 1;
    }

    # Set webhook
    $client->setWebhook($webhook) or do {
        $self->debug($client->trace);
        $self->debug($client->res->body) if $self->verbosemode;
        return $self->raise($client->error);
    };

    # Result
    my $result = $client->res->json("/description");
    if ($result) {
        say green($result)
    } else {
        say green("Done")
    }

    return 1;
});

__PACKAGE__->register_handler(
    handler     => "send",
    description => "Send message",
    code => sub {
### CODE:
    my ($self, $meta, @arguments) = @_;
    my $chatid = shift(@arguments);
    return $self->raise("No chatid specified") unless $chatid;
    my $_message = shift(@arguments) // "";
    my $message = decode(locale => $_message);
    return $self->raise("No message specified") unless length($message);
    my $client = $self->client;

    # Send
    $client->sendMessage(
        chat_id => $chatid,
        text => $message,
    ) or do {
        $self->debug($client->trace);
        $self->debug($client->res->body) if $self->verbosemode;
        return $self->raise($client->error);
    };

    # Result
    my $result = $client->res->json("/description");
    if ($result) {
        say green($result)
    } else {
        say green("Done")
    }

    return 1;
});

__PACKAGE__->register_handler(
    handler     => "photo",
    description => "Send photo/image/picture",
    code => sub {
### CODE:
    my ($self, $meta, @arguments) = @_;
    my $chatid = shift(@arguments);
    return $self->raise("No chatid specified") unless $chatid;
    my $_message = shift(@arguments) // "";
    my $message = decode(locale => $_message);
    return $self->raise("No caption message specified") unless length($message);
    my $_photo = shift(@arguments) // "";
    my $photo = length($_photo) ? glob(decode(locale => $_photo)) : '';
    return $self->raise("No path to photo specified or found") unless length($photo) && -f $photo;
    my $client = $self->client;

    # Send
    $client->sendPhoto(
        chat_id => $chatid,
        photo   => {
            file => encode(locale_fs => $photo),
        },
        caption => $message,
    ) or do {
        $self->debug($client->trace);
        $self->debug($client->res->body) if $self->verbosemode;
        return $self->raise($client->error);
    };

    # Result
    my $result = $client->res->json("/description");
    if ($result) {
        say green($result)
    } else {
        say green("Done")
    }

    return 1;
});

__PACKAGE__->register_handler(
    handler     => "video",
    description => "Send video",
    code => sub {
### CODE:
    my ($self, $meta, @arguments) = @_;
    my $chatid = shift(@arguments);
    return $self->raise("No chatid specified") unless $chatid;
    my $_message = shift(@arguments) // "";
    my $message = decode(locale => $_message);
    return $self->raise("No caption message specified") unless length($message);
    my $_video = shift(@arguments) // "";
    my $video = length($_video) ? glob(decode(locale => $_video)) : '';
    return $self->raise("No path to video specified or found") unless length($video) && -f $video;
    my $_thumbnail = shift(@arguments) // "";
    my $thumbnail = length($_thumbnail) ? glob(decode(locale => $_thumbnail)) : '';
    return $self->raise("No path to thumbnail specified or found") unless !length($thumbnail) || -f $thumbnail;
    my $client = $self->client;

    # Send
    $client->sendVideo(
        chat_id => $chatid,
        video => {
            file => encode(locale_fs => $video),
        },
        length($thumbnail) ? (
            thumbnail => {
                file => encode(locale_fs => $thumbnail),
            },
        ) : (),
        caption => $message,
    ) or do {
        $self->debug($client->trace);
        $self->debug($client->res->body) if $self->verbosemode;
        return $self->raise($client->error);
    };

    # Result
    my $result = $client->res->json("/description");
    if ($result) {
        say green($result)
    } else {
        say green("Done")
    }

    return 1;
});

__PACKAGE__->register_handler(
    handler     => "audio",
    description => "Send audio",
    code => sub {
### CODE:
    my ($self, $meta, @arguments) = @_;
    my $chatid = shift(@arguments);
    return $self->raise("No chatid specified") unless $chatid;
    my $_message = shift(@arguments) // "";
    my $message = decode(locale => $_message);
    return $self->raise("No caption message specified") unless length($message);
    my $_audio = shift(@arguments) // "";
    my $audio = length($_audio) ? glob(decode(locale => $_audio)) : '';
    return $self->raise("No path to audio specified or found") unless length($audio) && -f $audio;
    my $client = $self->client;

    # Send
    $client->sendAudio(
        chat_id => $chatid,
        audio   => {
            file => encode(locale_fs => $audio),
        },
        caption => $message,
    ) or do {
        $self->debug($client->trace);
        $self->debug($client->res->body) if $self->verbosemode;
        return $self->raise($client->error);
    };

    # Result
    my $result = $client->res->json("/description");
    if ($result) {
        say green($result)
    } else {
        say green("Done")
    }

    return 1;
});


__PACKAGE__->register_handler(
    handler     => "document",
    description => "Send file/docuemnt",
    code => sub {
### CODE:
    my ($self, $meta, @arguments) = @_;
    my $chatid = shift(@arguments);
    return $self->raise("No chatid specified") unless $chatid;
    my $_message = shift(@arguments) // "";
    my $message = decode(locale => $_message);
    return $self->raise("No caption message specified") unless length($message);
    my $_document = shift(@arguments) // "";
    my $document = length($_document) ? glob(decode(locale => $_document)) : '';
    return $self->raise("No path to document specified or found") unless length($document) && -f $document;
    my $_thumbnail = shift(@arguments) // "";
    my $thumbnail = length($_thumbnail) ? glob(decode(locale => $_thumbnail)) : '';
    return $self->raise("No path to thumbnail specified or found") unless !length($thumbnail) || -f $thumbnail;
    my $client = $self->client;

    # Send
    $client->sendDocument(
        chat_id => $chatid,
        document => {
            file => encode(locale_fs => $document),
        },
        length($thumbnail) ? (
            thumbnail => {
                file => encode(locale_fs => $thumbnail),
            },
        ) : (),
        caption => $message,
    ) or do {
        $self->debug($client->trace);
        $self->debug($client->res->body) if $self->verbosemode;
        return $self->raise($client->error);
    };

    # Result
    my $result = $client->res->json("/description");
    if ($result) {
        say green($result)
    } else {
        say green("Done")
    }

    return 1;
});


__PACKAGE__->register_handler(
    handler     => "server",
    description => "HTTP REST server",
    code => sub {
### CODE:
    my ($self, $meta, @arguments) = @_;

    # Dash k
    my $dash_k = shift(@arguments) || "status";
    unless (grep {$_ eq $dash_k} qw/start status stop restart reload/) {
        $self->error("Incorrect LSB command! Please use start, status, stop, restart or reload");
        return 0;
    }

    # Get permisions by names
    my $uid = $>; # Effect. UID
    my $gid = $); # Effect. GID
    if (IS_ROOT) {
        $uid = getpwnam(SERVER_USERNAME) || do {
            $self->error("getpwnam failed - $!");
            return 0;
        };
        $gid = getgrnam(SERVER_GROUPNAME) || do {
            $self->error("getgrnam failed - $!\n");
            return 0;
        };
    }

    # Prepare DataDir if not specified
    unless ($self->option("datadir")) {
        $self->datadir(path(sharedstatedir(), PROJECTNAMEL)->to_string()); # /var/lib/r2d2bot
        unless (-e $self->datadir) {
            unless (preparedir($self->datadir)) {
                $self->error(sprintf("Can't prepare directory %s", $self->datadir));
                return 0;
            }
        }
        # Set permisions (GID and UID) for work directory
        chown($uid, $gid, $self->datadir) if IS_ROOT && File::stat::stat($self->datadir)->uid != $uid;
    }

    # Prepare temp directory
    unless (-e $self->tempdir()) {
        unless (preparedir($self->tempdir, 0777)) {
            $self->error(sprintf("Can't prepare directory %s", $self->tempdir));
            return 0;
        }
    }

    # Prepare log directory
    if (IS_ROOT) {
        my $logdir = $self->logdir;
        unless (preparedir( $logdir, 0777 )) {
            $self->error(sprintf("Can't prepare log directory: %s", $logdir));
            return 0;
        }
        # Set permisions (GID and UID) for log directory
        chown($uid, $gid, $logdir) if File::stat::stat($logdir)->uid != $uid;
        $self->debug(sprintf("Log dir: %s", $self->logdir));
    } else {
        $self->logfile(path($self->tempdir(), sprintf("%s.log", PROJECTNAMEL))->to_string);
        $self->debug(sprintf("Log file: %s", $self->logfile));
    }

    # Prepare pid directory and file
    my $piddir = IS_ROOT ? path( rundir(), PROJECTNAMEL)->to_string : $self->tempdir();
    my $pidfile = path($piddir, sprintf("%s.pid", PROJECTNAMEL))->to_string;
    unless (preparedir($piddir)) {
        $self->error(sprintf("Can't prepare pid directory: %s", $piddir));
        return 0;
    }
    # Set permisions (GID and UID) for pid directory
    chown($uid, $gid, $piddir) if IS_ROOT && File::stat::stat($piddir)->uid != $uid;
    $self->debug(sprintf("Pid file: %s", $pidfile));

    # Hypnotoad variables
    my $upgrade = 0;
    my $reload = 0;
    my $upgrade_timeout = UPGRADE_TIMEOUT;

    # Mojolicious Application
    my $app = R2D2Bot::Server->new(ctk => $self);
       $app->attr(ctk => sub { $self }); # has ctk => sub { CTKx->instance->ctk };
    my $prefork = Mojo::Server::Prefork->new( app => $app ); # app => $self
       $prefork->pid_file($pidfile);

    # Hypnotoad Pre-fork settings
    $prefork->max_clients(tv2int(value($self->conf("clients")))) if defined $self->conf("clients");
    $prefork->max_requests(tv2int(value($self->conf("requests")))) if defined $self->conf("requests");
    $prefork->accepts(tv2int(value($self->conf("accepts")))) if defined $self->conf("accepts");
    $prefork->spare(tv2int(value($self->conf("spare")))) if defined $self->conf("spare");
    $prefork->workers(tv2int(value($self->conf("workers")))) if defined $self->conf("workers");

    # Make Listen
    my $cfg_listen = value($self->conf("listen"));
    my $tls_on = isTrueFlag(value($self->conf("tls")));
    my $listen = $tls_on ? "https://" : "http://";
    if ($cfg_listen) {
        $listen .= $cfg_listen;
    } else {
        $listen .= sprintf("%s:%d",
                value($self->conf("listenaddr")) || SERVER_LISTEN_ADDR,
                tv2int16(value($self->conf("listenport"))) || SERVER_LISTEN_PORT,
            );
    }
    my $_resolve_cf = sub {
        my $f = shift;
        return $f if File::Spec->file_name_is_absolute($f);
        return File::Spec->catfile($self->root, $f);
    };
    if ($tls_on) {
        my @p = ();
        foreach my $k (qw/ciphers version/) {
            my $v = value($self->conf("tls_$k")) // '';
            next unless length $v;
            push @p, sprintf("%s=%s", $k, $v);
        }
        foreach my $k (qw/ca cert key/) {
            my $v = value($self->conf("tls_$k")) // '';
            next unless length $v;
            push @p, sprintf("%s=%s", $k, $_resolve_cf->($v));
        }
        push @p, sprintf("%s=%s", "verify", value($self->conf("tls_verify")) || '0x00')
            if value($self->conf("tls_verify"));
        $listen .= sprintf("?%s", join('&', @p));
    }
    $prefork->listen([$listen]);

    # Working with Dash k
    if ($dash_k eq 'start') {
        if (my $pid = $prefork->check_pid()) {
            say "Already running $pid";
            return 1;
        }
    } elsif ($dash_k eq 'stop') {
        if (my $pid = $prefork->check_pid()) {
            kill 'QUIT', $pid;
            say "Stopping $pid";
        } else {
            say "Not running";
        }
        return 1;
    } elsif ($dash_k eq 'restart') {
        if (my $pid = $prefork->check_pid()) {
            $upgrade ||= steady_time;
            kill 'QUIT', $pid;
            my $up = $upgrade_timeout;
            while (kill 0, $pid) {
                $up--;
                sleep 1;
            }
            die("Can't stop $pid") if $up <= 0;
            say "Stopping $pid";
            $upgrade = 0;
        }
    } elsif ($dash_k eq 'reload') {
        my $pid = $prefork->check_pid();
        if ($pid) {
            # Start hot deployment
            kill 'USR2', $pid;
            say "Reloading $pid";
            return 1;
        }
        say "Not running";
    } else { # status
        if (my $pid = $prefork->check_pid()) {
            say "Running $pid";
        } else {
            say "Not running";
        }
        return 1;
    }

    #
    # LSB start
    #

    # This is a production server
    $ENV{MOJO_MODE} ||= 'production';

    # Listen USR2 (reload)
    $SIG{USR2} = sub { $upgrade ||= steady_time };

    # Set hooks
    #$prefork->on(spawn => sub () { # Spawn (start worker)
    #    my $self = shift; # Prefork object
    #    my $pid = shift;
    #    #say "Spawn (start) $pid";
    #    $self->app->log->debug("Spawn (start) $pid");
    #});
    $prefork->on(wait => sub { # Manage (every 1 sec)
        my $self = shift; # Prefork object

        # Upgrade
        if ($upgrade) {
            #$self->app->log->debug(">>> " . $self->healthy() || '?');
            unless ($reload) {
                $reload = 1; # Off next reloading
                if ($self->app->reload()) {
                    $reload = 0;
                    $upgrade = 0;
                    return;
                }
            }

            # Timeout
            if (($upgrade + $upgrade_timeout) <= steady_time()) {
                kill 'KILL', $$;
                $upgrade = 0;
            }
        }
    });
    #$prefork->on(reap => sub { # Cleanup (Emitted when a child process exited)
    #    my $self = shift; # Prefork object
    #    my $pid = shift;
    #    #say "Reap (cleanup) $pid";
    #    $self->app->log->debug("Reap (cleanup) $pid");
    #});
    $prefork->on(finish => sub { # Finish
        my $self = shift; # Prefork object
        my $graceful = shift;
        $self->app->log->debug($graceful ? 'Graceful server shutdown' : 'Server shutdown');
    });

    # Set GID and UID
    if (IS_ROOT) {
        if (defined($gid)) {
            POSIX::setgid($gid) || do {
                $self->error("setgid $gid failed - $!");
                return 0;
            };
            $) = "$gid $gid"; # this calls setgroups
            if (!($( eq "$gid $gid" && $) eq "$gid $gid")) { # just to be sure
                $self->error("detected strange gid");
                return 0;
            }
        }
        if (defined($uid)) {
            POSIX::setuid($uid) || do {
                $self->error("setuid $uid failed - $!");
                return 0;
            };
            if (!($< == $uid && $> == $uid)) { # just to be sure
                $self->error("detected strange uid");
                return 0;
            }
        }
    }

    # Daemonize
    $prefork->daemonize() unless $self->debugmode();

    # Running
    say "Running";
    $prefork->run();

    #my $fingerprint = $self->get_fingerprint;
    #say explain(\%exest);
    #my @strings = split("\n", $exest{output});
    #say explain(\@strings);

    return 1;
});

sub again {
    my $self = shift;
       $self->SUPER::again(); # CTK::App again first!!

    # Client instance
    $self->{client} = R2D2Bot::Client->new(
        token               => $self->conf("apitoken"),
        proxy               => $self->conf("proxy"),
        url_mask            => $self->conf("apiurl"),
        insecure            => $self->option("insecure"),
        max_redirects       => $self->conf("maxredirects"),
        connect_timeout     => $self->conf("connecttimeout"),
        inactivity_timeout  => $self->conf("inactivitytimeout"),
        request_timeout     => $self->conf("requesttimeout"),
    );

    # Cams
    my $cams = array($self->conf("cam"));
    $self->{_cams} = {_parse_cams($cams)};

    return $self; # CTK requires!
}
sub raise {
    my $self = shift;
    say STDERR red(@_);
    return 0;
}
sub client {
    my $self = shift;
    return $self->{client}
}
sub getcam {
    my $self = shift;
    my $cam = shift;
    my $cams = $self->{_cams};
    return unless is_hash($cams);
    return [keys %$cams] unless defined $cam;
    return unless $cams->{$cam};
    return array($cams, $cam);
}
sub execmd {
    my $self = shift;
    my @cmd = (@_);
    my $scmd = join(" ", @cmd);
    my $error;

    # Run command
    my $exe_err = '';
    my $exe_out = CTK::Util::execute([@cmd], undef, \$exe_err);
    my $stt = $? >> 8;
    my $exe_stt = $stt ? 0 : 1;
    chomp($exe_out) if defined($exe_out) && length($exe_out);
    if (!$exe_stt && $exe_err) {
        chomp($exe_err);
        $error = $exe_err;
    } elsif ($stt) {
        $error = sprintf("Exitval=%d", $stt);
    }

    return (
        command => $scmd,
        status  => $exe_stt,
        exitval => $stt,
        error   => $error,
        output  => $exe_out,
    );
}

# Helper functions
sub explain {
    my $dumper = Data::Dumper->new( [shift] );
    $dumper->Indent(1)->Terse(1);
    $dumper->Sortkeys(1) if $dumper->can("Sortkeys");
    return $dumper->Dump;
}

# Colored helper functions
sub green {  IS_TTY ? colored(['bright_green'],  sprintf(shift, @_)) : sprintf(shift, @_) }
sub red {    IS_TTY ? colored(['bright_red'],    sprintf(shift, @_)) : sprintf(shift, @_) }
sub yellow { IS_TTY ? colored(['bright_yellow'], sprintf(shift, @_)) : sprintf(shift, @_) }
sub cyan {   IS_TTY ? colored(['bright_cyan'],   sprintf(shift, @_)) : sprintf(shift, @_) }
sub blue {   IS_TTY ? colored(['bright_blue'],   sprintf(shift, @_)) : sprintf(shift, @_) }
sub magenta {IS_TTY ? colored(['bright_magenta'],sprintf(shift, @_)) : sprintf(shift, @_) }

# Internal functions
sub _parse_cams {
    my $in = shift || [];
    my %out = ();
    foreach my $cam (@$in) {
        my @wrds = shellwords($cam);
        my $k = shift(@wrds);
        next unless $k;
        next unless scalar(@wrds);
        $out{$k} = [@wrds];
    }
    return %out
}

1;

__END__
