package R2D2Bot::Client; # $Id: Client.pm 248 2021-11-16 17:26:22Z minus $
use strict;
use warnings FATAL => 'all';
use utf8;

=encoding utf8

=head1 NAME

R2D2Bot::Client - R2D2Bot web-client class

=head1 VERSION

Version 1.01

=head1 SYNOPSIS

    use R2D2Bot::Client;

    my $clinet = R2D2Bot::Client->new(
        token               => "1234567890:AAF0FJgBjxf8N7p1lanxNaibtGjCm0K763Q",
        max_redirects       => 2, # Default: 10
        connect_timeout     => 3, # Default: 10 sec
        inactivity_timeout  => 5, # Default: 30 sec
        request_timeout     => 10, # Default: 5 min (300 sec)
        insecure            => 0, # Default: 0 (on validation)
    );
    my $status = $client->check();

    if ($status) {
        print STDOUT $client->res->body;
    } else {
        print STDERR $clinet->error;
    }


=head1 DESCRIPTION

This module provides R2D2Bot web-client functionality

=head2 new

    my $clinet = R2D2Bot::Client->new(
        token               => "1234567890:AAF0FJgBjxf8N7p1lanxNaibtGjCm0K763Q",
        max_redirects       => 2, # Default: 10
        connect_timeout     => 3, # Default: 10 sec
        inactivity_timeout  => 5, # Default: 30 sec
        request_timeout     => 10, # Default: 5 min (300 sec)
        insecure            => 0, # Default: 0 (on validation)
    );

Returns client

=over 8

=item B<insecure>

The boolean mode of TLS validation. False - turns on the validation; True - turns off the validation

=item B<max_redirects>

Maximum number of redirects the user agent will follow before it fails. Default - 10

=item B<*timeout>

Timeout for connections, requests and inactivity periods in seconds.

=item B<token>

The API token

=item B<ua>

The Mojo UserAgent object

=item B<url_mask>

Full URL MASK of the WEB Server

Default: https://api.telegram.org/bot[TOKEN]/[METHOD]

=back

=head1 METHODS

=head2 check

    my $status = $client->check;

Returns check-status of server. 0 - Error; 1 - Ok

=head2 cleanup

    $client->cleanup;

Cleanup all variable data in object and returns client object

=head2 code

    my $code = $clinet->code;

Returns HTTP code of the response

=head2 error

    print $clinet->error;

Returns error string

=head2 req

    my $request = $clinet->req;

Returns Mojo::Message::Request object

=head2 request

    my $json = $clinet->request("METHOD", "API", ...ATTRIBUTES...);

Send request

=head2 res

    my $response = $clinet->res;

Returns Mojo::Message::Response object

=head2 sendMessage

    my $status = $client->sendMessage(
        chat_id => 1234567890,
        text    => 'Hello world!',
    );

Send message to chat

=head2 sendAudio

    my $status = $client->sendAudio(
        chat_id => 1234567890,
        photo   => {
            file => './src/img/music.mp3',
        },
        caption => 'Look at my cool music!'
    );

Send audio to chat

=head2 sendPhoto

    my $status = $client->sendPhoto(
        chat_id => 1234567890,
        photo   => {
            file => './src/img/droid.png',
        },
        caption => 'Look at my cool photo!'
    );

Send photo to chat

=head2 sendVideo

    my $status = $client->sendVideo(
        chat_id => 1234567890,
        thumbnail   => {
            file => './src/img/droid.png',
        },
        video   => {
            file => './src/img/droid.mp4',
        },
        caption => 'Look at my cool video!'
    );

Send video to chat

=head2 sendDocument

    my $status = $client->sendDocument(
        chat_id => 1234567890,
        thumbnail   => {
            file => './src/img/droid.png',
        },
        document   => {
            file => './src/img/droid.txt',
        },
        caption => 'My document!'
    );

Send file (document) to chat

=head2 setWebhook

    my $status = $client->setWebhook("https://example.com/trap")

Sets webhook

=head2 status

    my $status = $clinet->status;

Returns object status value. 0 - Error; 1 - Ok

=head2 trace

    my $trace = $client->trace;
    print $client->trace("New trace record");

Gets trace stack or pushes new trace record to trace stack

=head2 tx

    my $status = $clinet->tx($tx);

Works with Mojo::Transaction object, interface with it

=head2 ua

    my $ua = $clinet->ua;

Returns Mojo::UserAgent object

=head2 url

    my $url_object = $clinet->url;

Returns Mojo::URL object


=head1 SEE ALSO

L<Mojolicious>, L<R2D2Bot>, L<Mojo::UserAgent>, L<WWW::Telegram::BotAPI>

=head1 AUTHOR

Serż Minus (Sergey Lepenkov) L<http://www.serzik.com> E<lt>minus@serzik.comE<gt>

=head1 COPYRIGHT

Copyright (C) 1998-2021 D&D Corporation. All Rights Reserved

=head1 LICENSE

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

See C<LICENSE> file

=cut

use CTK::Util qw/dformat dtf lf_normalize/;
use Mojo::URL;
use Mojo::UserAgent;
use Mojo::UserAgent::Proxy;
use Mojo::Util qw/steady_time/;

use vars qw/ $VERSION /;
$VERSION = "1.01";

use constant {
        API_URL_MASK        => "https://api.telegram.org/bot[TOKEN]/[METHOD]",
        MAX_REDIRECTS       => 10,
        CONNECT_TIMEOUT     => 10,
        INACTIVITY_TIMEOUT  => 30,
        REQUEST_TIMEOUT     => 180,
        TRANSACTION_MASK    => "%s %s >>> %s %s [%s in %s%s]", # GET /auth >>> 200 OK [1.04 KB in 0.0242 seconds (43.1 KB/sec)]
        CONTENT_TYPE        => "application/json",
    };

sub new {
    my $class = shift;
    my %args  = @_;

    # General
    $args{url_mask} ||= API_URL_MASK; # URL Mask
    $args{token} ||= "";
    if ($args{url}) {
        $args{url} = Mojo::URL->new($args{url});
    } else {
        $args{url} = Mojo::URL->new(dformat($args{url_mask}, {
            TOKEN => $args{token},
        }));
    }
    $args{prefix} = $args{url}->path->{path} // "";
    $args{proxy} //= "";
    $args{status} = 1; # 0 - error, 1 - ok
    $args{error} = ""; # string
    $args{code} = 0;   # integer
    $args{trace} = []; # trace pool
    $args{tx_time} = 0;
    $args{req} = undef;
    $args{res} = undef;

    # User Agent
    my $ua = $args{ua};
    unless ($ua) {
        # Create the instance
        $ua = Mojo::UserAgent->new(
                max_redirects       => $args{max_redirects} || MAX_REDIRECTS,
                connect_timeout     => $args{connect_timeout} || CONNECT_TIMEOUT,
                inactivity_timeout  => $args{inactivity_timeout} || INACTIVITY_TIMEOUT,
                request_timeout     => $args{request_timeout} || REQUEST_TIMEOUT,
                insecure            => $args{insecure} || 0,
            );
        $ua->transactor->name(sprintf("%s/%s", __PACKAGE__, __PACKAGE__->VERSION));

        # Set proxy
        my $proxy = Mojo::UserAgent::Proxy->new;
        $ua->proxy($args{url}->scheme =~ /s$/
            ? $proxy->https($args{proxy})
            : $proxy->http($args{proxy})) if $args{proxy};
        $args{ua} = $ua;
    }

    my $self = bless {%args}, $class;
    return $self;
}
sub error {
    my $self = shift;
    my $e = shift;
    $self->{error} = $e if defined $e;
    return $self->{error};
}
sub status {
    my $self = shift;
    my $s = shift;
    $self->{status} = $s if defined $s;
    return $self->{status};
}
sub code {
    my $self = shift;
    my $c = shift;
    $self->{code} = $c if defined $c;
    return $self->{code};
}
sub trace {
    my $self = shift;
    my $v = shift;
    if (defined($v)) {
        my $a = $self->{trace};
        push @$a, lf_normalize($v);
        return lf_normalize($v);
    }
    my $trace = $self->{trace} || [];
    return join("\n",@$trace);
}
sub cleanup {
    my $self = shift;
    $self->{status} = 1;
    $self->{error} = "";
    $self->{code} = 0;
    undef $self->{req};
    $self->{req} = undef;
    undef $self->{res};
    $self->{res} = undef;
    undef $self->{trace};
    $self->{trace} = [];
    return $self;
}
sub req {
    my $self = shift;
    return $self->{req};
}
sub res {
    my $self = shift;
    return $self->{res};
}
sub url {
    my $self = shift;
    return $self->{url};
}
sub ua {
    my $self = shift;
    return $self->{ua};
}
sub tx {
    my $self = shift;
    my $tx = shift;

    # Check Error
    my $err = $tx->error;
    unless (!$err || $err->{code}) {
        $self->error($err->{message});
        $self->status(0);
    }
    $self->code($tx->res->code || "000");
    $self->status($tx->res->is_success ? 1 : 0);
    $self->error($tx->res->json("/description") || $tx->res->json("/message") || $err->{message} || "Unknown error" )
        if $tx->res->is_error && !$self->error;

    # Tracing
    my $length = $tx->res->body_size || 0;
    my $rtime = $self->{tx_time} // 0;
    $self->trace(sprintf(TRANSACTION_MASK,
        $tx->req->method, $tx->req->url->to_abs, # Method & URL
        $self->code, $tx->res->message || $err->{message} || "Unknown error", # Line
        _fbytes($length), # Length
        _fduration($rtime), # Duration
        $rtime ? sprintf(" (%s/sec)", _fbytes($length/$rtime)) : "",
      ));
    my $req_hdrs = $tx->req->headers->to_string;
    if ($req_hdrs) {
        $self->trace(join("\n", map {$_ = "> $_"} split(/\n/, $req_hdrs)));
        $self->trace(">");
    }
    my $res_hdrs = $tx->res->headers->to_string;
    if ($res_hdrs) {
        $self->trace(join("\n", map {$_ = "< $_"} split(/\n/, $res_hdrs)));
        $self->trace("<");
    }

    # Request And Response
    $self->{req} = $tx->req;
    $self->{res} = $tx->res;

    return $self->status;
}
sub request {
    my $self = shift;
    my $meth = shift; # HTTP Method
    my $api_meth = shift; # API Method
    my @params = @_;
    $self->cleanup(); # Cleanup first

    # Correct api meth
    my $url = $self->url->clone;
    $url->path(dformat($self->{prefix}, {
        METHOD => $api_meth,
    }));
    #return $url->to_string;

    # Request
    my $start_time = steady_time() * 1;
    my $tx = $self->ua->build_tx($meth, $url, @params); # Create transaction (tx)
    my $status = $self->tx($self->ua->start($tx)); # Run it and validate!);
    $self->{tx_time} = sprintf("%.*f",4, steady_time()*1 - $start_time) * 1;

    return $status;
}
sub check {
    my $self = shift;
    my $meth = shift // "getMe";
    return $self->request(GET => $meth, {accept => CONTENT_TYPE}); # "*/*"
}
sub sendMessage {
    my $self = shift;
    my %params = @_;
    return $self->request(POST => "sendMessage",
        { # Headers
            accept => CONTENT_TYPE, # "*/*"
        },
        json => {%params},
    );
}
sub sendPhoto {
    my $self = shift;
    my %params = @_;
    return $self->request(POST => "sendPhoto",
        { # Headers
            accept => CONTENT_TYPE, # "*/*"
            'Content-Type' => 'multipart/form-data',
        },
        form => {%params},
    );
}
sub sendAudio {
    my $self = shift;
    my %params = @_;
    return $self->request(POST => "sendAudio",
        { # Headers
            accept => CONTENT_TYPE, # "*/*"
            'Content-Type' => 'multipart/form-data',
        },
        form => {%params},
    );
}
sub sendVideo {
    my $self = shift;
    my %params = @_;
    return $self->request(POST => "sendVideo",
        { # Headers
            accept => CONTENT_TYPE, # "*/*"
            'Content-Type' => 'multipart/form-data',
        },
        form => {%params},
    );
}
sub sendDocument {
    my $self = shift;
    my %params = @_;
    return $self->request(POST => "sendDocument",
        { # Headers
            accept => CONTENT_TYPE, # "*/*"
            'Content-Type' => 'multipart/form-data',
        },
        form => {%params},
    );
}
sub setWebhook {
    my $self = shift;
    my $url = shift;
    unless ($url) {
        $self->error("No webhook URL");
        return $self->status(0);
    }

    # Request
    return $self->request(POST => "setWebhook",
        { # Headers
            accept => CONTENT_TYPE, # "*/*"
        },
        json => {
            url => $url,
        },
    );
}

sub _fduration {
    my $msecs = shift || 0;
    my $secs = int($msecs);
    my $hours = int($secs / (60*60));
    $secs -= $hours * 60*60;
    my $mins = int($secs / 60);
    $secs %= 60;
    if ($hours) {
        return sprintf("%d hours %d minutes", $hours, $mins);
    } elsif ($mins >= 2) {
        return sprintf("%d minutes", $mins);
    } elsif ($secs < 2*60) {
        return sprintf("%.4f seconds", $msecs);
    } else {
        $secs += $mins * 60;
        return sprintf("%d seconds", $secs);
    }
}
sub _fbytes {
    my $n = int(shift);
    if ($n >= 1024 * 1024) {
        return sprintf "%.3g MB", $n / (1024.0 * 1024);
    } elsif ($n >= 1024) {
        return sprintf "%.3g KB", $n / 1024.0;
    } else {
        return "$n bytes";
    }
}

1;

__END__
