package R2D2Bot::Server; # $Id: Server.pm 255 2021-11-17 10:41:51Z minus $
use strict;
use warnings FATAL => 'all';
use utf8;

=encoding utf8

=head1 NAME

R2D2Bot::Server - R2D2Bot web-server class

=head1 VERSION

Version 1.00

=head1 SYNOPSIS

    use R2D2Bot::Server;

=head1 DESCRIPTION

This module provides R2D2Bot web-server functionality

=head2 reload

The reload hook

=head2 startup

Mojo application startup method

=head1 SEE ALSO

L<Mojolicious>, L<R2D2Bot>

=head1 AUTHOR

Serż Minus (Sergey Lepenkov) L<http://www.serzik.com> E<lt>minus@serzik.comE<gt>

=head1 COPYRIGHT

Copyright (C) 1998-2021 D&D Corporation. All Rights Reserved

=head1 LICENSE

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

See C<LICENSE> file

=cut

use vars qw/ $VERSION /;
$VERSION = "1.00";

use Mojo::Base 'Mojolicious';

use Mojo::File qw/path/;
use Mojo::Util qw/sha1_sum secure_compare/;

use CTK::Util qw/preparedir sharedir sharedstatedir/;
use CTK::ConfGenUtil;


has 'documentroot';

# Usage page (MarkdownV2)
#
# *bold \*text*
# _italic \*text_
# __underline__
# ~strikethrough~
# *bold _italic bold ~italic bold strikethrough~ __underline italic bold___ bold*
# [inline URL](http://www.example.com/)
# [inline mention of a user](tg://user?id=123456789)
# `inline fixed-width code`
# ```
# pre-formatted fixed-width code block
# ```
# ```python
# pre-formatted fixed-width code block written in the Python programming language
# ```
#
use constant USAGE => <<'USAGE';
*General commands:*
/help, /usage \- show this page
/status \- show welcome string

*Camera commands:*
/cam \- get list of cameras
/cam_NAME_ \- get snapshot drom camera
USAGE

# Mojo Routes (paths)
sub startup {
    my $self = shift; # app
    my $ctk = $self->can('ctk') ? $self->ctk() : $self->{ctk};

    # Set password
    my $username = value($ctk->conf("username") // "");
    my $secret = value($ctk->conf("password") // "");
    $self->secrets([$secret]) if length($secret);

    # Logging
    $self->log->level($ctk->debugmode ? "debug" : "warn")->path($ctk->logfile());
    #$self->log->debug("Startup!! =$$"); # $self->ctk->logdir()

    # Switch to installable home directory
    $self->home(Mojo::Home->new($ctk->datadir()));

    # Get DocumentRoot and replace as public-path
    my $documentroot = value($ctk->conf("documentroot")) || path(sharedir(), $ctk->prefix)->to_string();
    $self->documentroot($documentroot);
    $self->static->paths()->[0] = $documentroot; #unshift @{$static->paths}, '/home/sri/themes/blue/public';
    $self->static->paths()->[1] = $ctk->datadir();

    # Hooks
    $self->hook(before_dispatch => sub {
        my $c = shift;

        # Set Server header
        $c->res->headers->server(sprintf("%s/%s", R2D2Bot::PROJECTNAME(), R2D2Bot->VERSION));

        # Parse URL path
        #my %norfcnum = $c->app->ctk->parse_norfcnum($c->req->url->path);
        #$c->app->log->info(" !!!! " . $need_auth); # $self->ctk->logdir()

        # Authentication
        my $need_auth = length($username) ? 1 : 0;
        if ($need_auth) {
            my $req_uri = $c->req->url->to_abs();
            my $ui_username = $req_uri->username() // "";
            my $ui_secret = sha1_sum($req_uri->password() // time());

            # Check username and password
            return 1 if length($secret)
                and secure_compare($username, $ui_username)
                and secure_compare($secret, $ui_secret);

            # Require authentication
            $c->res->headers->www_authenticate('Basic realm="R2D2Bot Strict Zone"');
            return $c->render(json => {
                message => "Authentication required!",
            }, status => 401);
        }

        return;
    });

    # Routes
    $self->routes->get('/')->to('alpha#root');
    $self->routes->get('/env')->to('alpha#env') if $ctk->debugmode();
    $self->routes->get('/info')->to('alpha#info');
    $self->routes->post('/send/:chatid' => [chatid => qr/[a-z0-9]+/])->to('alpha#sendm');
    $self->routes->post('/trap')->to('alpha#trap'); # From telegram
    $self->routes->post('/hook')->to('alpha#trap'); # ...alias

    # Favicon
    delete $self->static->extra->{'favicon.ico'};

    return 1;
}

# Reload hook
sub reload {
    my $self = shift;
    $self->log->warn("Request for reload $$"); # $self->ctk->logdir()
    return 1; # 1 - ok; 0 - error :(
}

1;

# Controller
package R2D2Bot::Server::Alpha;
use utf8;

use Mojo::Base 'Mojolicious::Controller';

use Mojo::File qw/path/;
#use Mojo::JSON qw/from_json/;
use Mojo::Util qw/b64_decode/;

use Try::Tiny;

use CTK::Util qw/sharedir sharedstatedir trim dformat dtf/;
use CTK::ConfGenUtil;

use vars qw/ $VERSION /;
$VERSION = "1.00";

use constant {
        CONTENT_TYPE_JSON   => "application/json",
        CONTENT_TYPE_TXT    => "text/plain; charset=utf-8",
        DATETIME_GMT_FORMAT => "%YYYY-%MM-%DD %hh:%mm:%ss %G",
    };

sub root {
    my $self = shift;
    return $self->reply->static('index.html');
}
sub env {
    my $self = shift;
    $self->res->headers->content_type(CONTENT_TYPE_TXT);
    return $self->render(text => $self->dumper(\%ENV));
}
sub info {
    my $self = shift; # Mojolicious::Routes
    my $ctk = $self->app->ctk();
    my $client = $ctk->client;

    # Logging
    #$self->app->log->info("Catched! Info log mark") if $ctk->verbosemode();

    # Response
    $self->res->headers->content_type(CONTENT_TYPE_JSON);
    return $self->render(json => {
            ok => \1,
            message => "Ok",
            generated => dtf(DATETIME_GMT_FORMAT, time(), 1),
            status => $ctk->status,
            $ctk->debugmode() || $ctk->verbosemode() ? (
                mode => $ctk->debugmode()
                    ? "debug mode"
                    : $ctk->verbosemode()
                        ? "verbose mode"
                        : "Oops!",
                datadir => $ctk->datadir,
                homedir => $self->app->home(),
                documentroot => $self->app->documentroot,
                tempdir => $ctk->tempdir,
                logdir => $ctk->logdir,
                logfile => $ctk->logfile,
                paths => $self->app->static->paths(),
                cams => $ctk->getcam(),

                # Client
                client => $client->status ? "ok" : $client->error,

            ) : (),
        });
}
sub sendm {
    my $self = shift;
    my $ctk = $self->app->ctk();
    my $chatid = $self->param("chatid");
    #my $raw = $self->req->body // '';

    # Response
    $self->res->headers->content_type(CONTENT_TYPE_JSON);
    return $self->render(json => {
        ok => \0,
        message => "This method is not implemented",
        generated => dtf(DATETIME_GMT_FORMAT, time(), 1),
        chatid => $chatid,
    }, status => 501);
}
sub trap {
    my $self = shift;
    my $ctk = $self->app->ctk();
    my $client = $ctk->client;

    # Data from telegram
    my $message = trim($self->req->json("/message/text") // "");
    my $chat_id = $self->req->json("/message/chat/id") // "";
    my $username = $self->req->json("/message/from/username") // "";
    my $first_name = $self->req->json("/message/from/first_name") // "";

    #$self->app->log->debug(sprintf("Incoming message: %s", $message ));
    #$self->app->log->debug(R2D2Bot::explain($self->req->json()));

    # Response
    $self->res->headers->content_type(CONTENT_TYPE_JSON);

    # Check commands
    if ($message =~ /^\//) {
        $message =~ s/^\/+//;
        $message = lc($message);
        $self->app->log->debug(sprintf("Command: %s", $message)) if $message;
    } else {
        return $self->render(json => {
            ok => \0,
            message => "Incorrect command",
            generated => dtf(DATETIME_GMT_FORMAT, time(), 1),
        });
    }

    # Main switch
    if ($message eq 'help' or $message eq 'usage') {
        $client->sendMessage(
            chat_id => $chat_id,
            parse_mode => "MarkdownV2",
            text => R2D2Bot::Server::USAGE(),
        ) or do {
            $self->app->log->error($client->error);
            $self->app->log->debug($client->trace) if $ctk->verbosemode;
        };
    } elsif ($message eq 'start') {
        # Prepare output
        my @text = ();
        push @text, sprintf("Hello %s!", $first_name || sprintf("@%s", $username));
        push @text, sprintf("Welcome to %s", R2D2Bot::PROJECTNAME());
        push @text, "See help page: /help";

        $client->sendMessage(
            chat_id => $chat_id,
            text => join("\n", @text),
        ) or do {
            $self->app->log->error($client->error);
            $self->app->log->debug($client->trace) if $ctk->verbosemode;
        };
    } elsif ($message eq 'status') {
        unless ($client->check()) {
            $self->app->log->error($client->error);
            return $self->render(json => {
                ok => \0,
                message => $client->error,
                generated => dtf(DATETIME_GMT_FORMAT, time(), 1),
            });
        }

        # Prepare output
        my @text = ();
        push @text, sprintf("Hello %s!", $first_name || sprintf("@%s", $username));
        push @text, sprintf("Bot: @%s", $client->res->json("/result/username") || R2D2Bot::PROJECTNAME());
        push @text, sprintf("ChatID: %s", $chat_id);

        # Send message
        $client->sendMessage(
            chat_id => $chat_id,
            text => join("\n", @text),
        ) or do {
            $self->app->log->error($client->error);
            $self->app->log->debug($client->trace) if $ctk->verbosemode;
        };
    } elsif ($message eq 'cam') {
        my $cams = $ctk->getcam();
        if (@$cams) {
            my @buttons;
            foreach my $cam (@$cams) {
                push @buttons, {
                    text => sprintf("/cam%s", $cam),
                };
            }
            $client->sendMessage(
                chat_id => $chat_id,
                text => sprintf('Available cameras: %s', join(", ", @$cams)),
                reply_markup => { # Object: ReplyKeyboardMarkup
                    resize_keyboard => \1,
                    keyboard => [[@buttons]]
                },
            ) or do {
                $self->app->log->error($client->error);
                $self->app->log->debug($client->trace) if $ctk->verbosemode;
            };
        } else {
            $client->sendMessage(
                chat_id => $chat_id,
                text => 'No available cameras',
            ) or do {
                $self->app->log->error($client->error);
                $self->app->log->debug($client->trace) if $ctk->verbosemode;
            };
        }
    } elsif ($message =~ /cam(.+)/) {
        my $cam = $1;
        my $cmd = $ctk->getcam($cam);
        if ($cmd && is_array($cmd)) {
            my %exest = $ctk->execmd(@$cmd);
            my $file;
            for (qw/jpg jpeg png gif bin/) {
                $file = path($ctk->tempdir, sprintf("%s.%s", $cam, $_))->to_string;
                last if -e $file;
            }
            if ($exest{status} && -f $file) {
                $client->sendPhoto(
                    chat_id => $chat_id,
                    photo   => {
                        file => $file,
                    },
                    caption => sprintf("Camera %s on %s", $cam, dtf(DATETIME_GMT_FORMAT, time(), 1)),
                ) or do {
                    $self->app->log->error($client->error);
                    $self->app->log->debug($client->trace) if $ctk->verbosemode;
                };
            } else {
                $exest{error} ||= "Can't get camera snapshot";
                $self->app->log->error($exest{error});
                $client->sendMessage(
                    chat_id => $chat_id,
                    text => $exest{error},
                ) or do {
                    $self->app->log->error($client->error);
                    $self->app->log->debug($client->trace) if $ctk->verbosemode;
                };
            }
        } else {
            $client->sendMessage(
                chat_id => $chat_id,
                text => 'Camera not found. Try /cam first',
            ) or do {
                $self->app->log->error($client->error);
                $self->app->log->debug($client->trace) if $ctk->verbosemode;
            };
        }
    } else { # Incorrect command
        $client->sendMessage(
            chat_id => $chat_id,
            parse_mode => "MarkdownV2",
            text => sprintf("Incorrect command ~%s~\\!\nSee help page: /help", $message),
        ) or do {
            $self->app->log->error($client->error);
            $self->app->log->debug($client->trace) if $ctk->verbosemode;
        };
    }

    # Response
    return $self->render(json => {
        ok => \1,
        message => "Ok",
        generated => dtf(DATETIME_GMT_FORMAT, time(), 1),
    });
}

1;

__END__
